#include "Cylinder.h"


//******************************************************************
// Default constructor.
//******************************************************************
Cylinder::Cylinder() : Shape(CYLINDER), center(0, 0), radius(0)
{}

//******************************************************************
// Constructor.
//******************************************************************
Cylinder::Cylinder(Point c, double r) : Shape(CYLINDER)
{
	Point llc(c.getX() - r, c.getY() - r); 
	lowerLeftCorner = llc;
	Shape::setWidth(2.0*r);
	Shape::setHeight(2.0*r);
	radius = r;
	center = c;
}


//******************************************************************
// Copy constructor
//******************************************************************
Cylinder::Cylinder(const Cylinder & C)
{
	int i;

	lowerLeftCorner = C.getCornerPoint();
	width = C.getWidth(); height = C.getHeight();
	type = C.getType(); 
	center = C.getCenter(); radius = C.getRadius();
}


//******************************************************************
// overloaded "=" operator
//******************************************************************
Cylinder & Cylinder::operator = (const Cylinder & C)
{
	int i;

	// A check to make sure we're not trying to copy ourself to
	// ourself. If this did happen, then we would delete all
	// information before we had a chance to copy it.
	if(this == &C)
		return *this;

	radius = C.getRadius();
	center = C.getCenter();
	width = C.getWidth();
	height = C.getHeight();
	type = C.getType(); 
	lowerLeftCorner = C.getCornerPoint(); 
	
	return (*this);
}

//******************************************************************
//  overloaded ostream operator (standard output)
//******************************************************************
ostream & operator << ( ostream & str, const Cylinder & C)
{
	str << "radius = " << C.radius << endl;
	str << "center = " << C.center << endl;
	return str;
}	


//******************************************************************
// overloaded ofstream operator (file output)
//******************************************************************
ofstream & operator << ( ofstream & fStr, const Cylinder & C)
{
	fStr << "radius = " << C.radius << endl;
	fStr << "center = " << C.center << endl;
	return fStr;
}	





//******************************************************************
//  Keeps other variables consistent when one is modified.
//******************************************************************
void Cylinder::update(VARTYPE vtype)
{
	switch (vtype) {
		case RADIUS:
			width = 2.0 * radius;
			height = 2.0 * radius;
			lowerLeftCorner.setXY(center.getX() + width / 2.0,
			                         center.getY() + height / 2.0);
			break;
		case CENTER:
			lowerLeftCorner.setXY(center.getX() + width / 2.0,
			                         center.getY() + height / 2.0);
			break;
		case WIDTH:
			radius = width / 2.0;
			height = width;
			center.setXY(lowerLeftCorner.getX() - width / 2.0,
			                lowerLeftCorner.getY() - height / 2.0);
			break;
		case HEIGHT:
			radius = height / 2.0;
			width = height;
			center.setXY(lowerLeftCorner.getX() - width / 2.0,
			                lowerLeftCorner.getY() - height / 2.0);
			break;
		case ULC:
			center.setXY(lowerLeftCorner.getX() + width / 2.0,
			                lowerLeftCorner.getY() + height / 2.0);
			break;
		default:
			cout << "error in setting a value in Cylinder\n";
			
	}
}			



//******************************************************************
// returns the area of a circle.
//******************************************************************
double Cylinder::area() const
{ 
	return M_PI * radius*radius;
}

//******************************************************************
// Determines if the point defined by (x, y) is inside of our
// shape, or at least within a certain proximity of it.
//******************************************************************
int Cylinder::isInside(double x, double y)
{
	double xC = center.getX();
	double yC = center.getY();
	double rhs;

	// We use the equation for a circle to test if the point
	// is inside. r^2 = (x - h)^2 + (y - k)^2
	rhs = (x  - xC) * (x - xC) + (y  - yC) * (y - yC);
		  
	if(rhs <= (radius*radius))
		return 1;
	else
		return 0;
}

//******************************************************************
// Calculates the ratio of the distance of particle "j" to a tangent
// line on the surface and the distance of particle "i" to this same
// line.
//******************************************************************
double Cylinder::calcRatioOfDistances(Point iPos, Point jPos)
{
	double di, dj, x1, y1, x1Plus, x1Minus, h, k, x2, y2;
	double i_x, i_y, j_x, j_y;
	double m1 = 0, m2;
	
	i_x = iPos.getX();
	i_y = iPos.getY();
	j_x = jPos.getX();
	j_y = jPos.getY();
	
	h = center.getX();
	k = center.getY();
	
	// First find the point (x1, y1) that is the intersection of a
	// line consisiting of the center of this circle (h, k) and the
	// position of particle i (i_x, i_y) AND the circle given by
	// center (h, k) and radius 'radius'.
	// We must be careful with vertical lines.
	if((i_x - h) != 0)
	{
		m1 = (i_y - k) / (i_x - h);
		x1Plus = (radius*radius) / (1+m1*m1) + h;
		x1Minus = -1.0 * (radius*radius) / (1+m1*m1) + h;
		
		// We need to figure out which intersecting point is the 
		// one we are interested in.
		if(h < i_x)
		{
			if(h <= x1Plus && x1Plus <= i_x)
				x1 = x1Plus;
			else
				x1 = x1Minus;
		} else {
			if(i_x <= x1Plus && x1Plus <= h)
				x1 = x1Plus;
			else
				x1 = x1Minus;
		}

		y1 = m1*(x1 - h) + k;
		
	} else {
	
		x1 = i_x;

		// We need to figure out which intersecting point is the 
		// one we are interested in.
		if(k < i_y)
			y1 = k + radius;
		else 
			y1 = k - radius;
	}
	
	
	// We now have the point on the surface of the circle that
	// represents the tangent point (x1, y1). Next, we need to calculate
	// the point at which the distance from this line to the boundary
	// point is the minimum (x2, y2).
	if (m1 != 0.0)
	{
		m2 = -1.0 / m1;
		x2 =(j_y - y1 + m2*x1-m1*j_x) / (m2 - m1);
		y2 = m1*(x2 -j_x) + j_y;
		
		dj = sqrt((j_x - x2)*(j_x - x2) + (j_y - y2)*(j_y - y2));
		di = sqrt((i_x - x1)*(i_x - x1) + (i_y - y1)*(i_y - y1));
		 
		return (dj / di);
	} else {
		dj = fabs(j_y - y1);
		di = fabs(i_y - y1);
			
		return (dj / di);
	}
}	
				
		
