//***********************************************************
//
// This is a base class for SlaveGrid. It contains the shapes
// and particles and methods to modify them, but does not do 
// anything with the simulation
//
// Written by: Brian Schlatter
// MOdified by: Keith Vertanen
// Last updated: 6/12/1999
//
//********************************************************** 
#ifndef GRID_HPP
#define GRID_HPP

#include <stdlib.h>
#include <assert.h>

#include "Parameter.h"
#include "Cell.h"
#include "Point.h"
#include "SmoothParticle.h"
#include "Shape.h"
#include "Box.h"
#include "Cylinder.h"
#include "FastLinkedList.h"
#include "msgtags.h"
#include "Timer.h"
#include <iomanip.h>
#include <iostream.h>
#include <math.h>

class Grid {

	public:
		Grid();
		Grid(Parameter );
		Grid(const Grid & G);
		~Grid();
		
		// accessor methods
		double getCellSize()    const { return cellSize;      }
		int getNumCellsWide()   const { return numCellsWide;  }
		int getNumCellsHigh()   const { return numCellsHigh;  }
		int getStartCellY()     const { return startCellY;    }
		int getEndCellY()       const { return endCellY;      }
		int getNumPts()         const { return numPts;        }
		int getEndOfPts()       const { return endOfPts;      }
		int getNumShapePts()    const { return numShapePts;   }
		int getCurNumShapePts() const { return curNumShapePts;}
		int getNumShapes()      const { return numShapes;      }
		Shape * getShape(int i) const { return theShapes[i];   }
		double getContainerHeight() const { return containerHeight;}
		double getContainerWidth()  const { return containerWidth; }
		int getFrame(float **);		
		
		// mutators
		void setNumShapePts(int nsp);
		void setNumShapes(int ns);
		void setNumPts(int np); 
		void setNumCellsWide(int wide);
		void setNumCellsHigh(int high);
		void updateNumShapePts(int nsp);
		void setCellSize(double cs)		   { cellSize = cs;	       }
		void setParam(Parameter p)		   { param = p;            }
		void setContainerHeight(double ch) { containerHeight = ch; }
		void setContainerWidth(double cw)  { containerWidth = cw;  }
		
		void addShape(double [] );
		void getShapeAttributes(int, double []);
	
		void addNewFreeParticle(Point, int );
		void addNewBndParticle(int, Point, int, int );
		void addNewMigrantParticle(SmoothParticle, SmoothParticle, SmoothParticle);
		void buildGrid();
		void calcCellSize();
		void calcNumCellsWideHigh(double, double);
		void setRangeMaster();

		// return a particle at a particular point in the array
		SmoothParticle getParticle(int);
		SmoothParticle getParticleID(int);
		SmoothParticle getShapeParticle(int);
		SmoothParticle getPlusHalfParticle(int);
		SmoothParticle getMinusHalfParticle(int);

		void deleteAllFree();		// delete all the free particles in our arrays

		int hashI(double );		// find the cell column given a x coordinate
		int hashJ(double );		// find the cell row given a y coordinate


		// file output rountines
		void writeID(ofstream &);	// write to debug file with IDs and all physical attributes
		void writeVerbose(ofstream &);  // write out to file with all particle information
		void writeNeighbor(ofstream &);  

		void printVerbose();		// debug, prints out contents of the grid
		void printMinMaxDensity();

		Grid & operator = (const Grid & );
		friend ostream & operator << ( ostream & , const Grid & );
		friend ofstream & operator << ( ofstream & , const Grid & );

		Parameter param;		// holds the simulation parameters

		int startCellY;
		int endCellY;

		long debugTimeSteps;

		Shape ** theShapes;
		int numShapes;
		int curShapes;
		
	protected:
		void deletePt(int i);		// delete the particle at index position i
		void error(char *);
		int  getFreeIndex();		// returns the an index to a free spot in the particle arrays
		void updateEndOfPts();		// move the endOfPts index to the left if possible

		int endOfPts;			// the next free index in the array (if freeSpace is empty)
		FastLinkedList freeSpace;	// tracks empty space in our particle array

		double cellSize;
		int numCellsWide;
		int numCellsHigh;		// equals endCellY - startCellY
		int numPts;
		int numShapePts;
		int curNumShapePts;

		double containerWidth;
		double containerHeight;

		Cell ** cells;			// 2D array of cells (linked list with array indices)

		// arrays that hold the particle and shape data
		SmoothParticle * t;
		SmoothParticle * tPlusOneHalf;
		SmoothParticle * tMinusOneHalf;

		SmoothParticle * shapePtsT;
		SmoothParticle * shapePtsTMinusOneHalf;
		SmoothParticle * shapePtsTPlusOneHalf;

};

#endif
