
#include "SlaveMPI.h"
#include "mpi.h"

//******************************************************************
// Construct a new SlaveMPI object, receive our rank, our north
// neighbor's rank, south neighbor's rank, and whether debugging 
// is on.
//******************************************************************
SlaveMPI::SlaveMPI(int r, int nr, int sr, int debug)
{
	if (debug>2)
		DEBUG = 1;
	else
		DEBUG = 0;

	our_rank   = r;
	north_rank = nr;
	south_rank = sr;

	updateTag = UPDATE_OFFSET;
}

//******************************************************************
// Receive a command from the master, return the result
//******************************************************************
int SlaveMPI::getCommand()
{
	int command;
	MPI_Status status;

	// receive a command from our master
	MPI_Recv(&command,1,MPI_INT,0,COMMAND,MPI_COMM_WORLD,&status);

	if (DEBUG) printf("Slave recv'd a command = %d\n", command);

	return command;
}

//******************************************************************
// Perform a performance benchmark test on ourself, send the 
// performance number and our system name to the master
//******************************************************************
void SlaveMPI::runStats()
{
	// master wants our name and mflops speed rating
        
	// run the timing function
	double flops = 0.0;
	Bench bench;        

	flops = bench.measureFlops();

	// get the name of our computer
        char hostname[80];

	// take no chances
	for (int i=0; i<80; i++)
		hostname[i] = '\0';

    gethostname(hostname, 80);

    MPI_Send(&flops, 1, MPI_DOUBLE, 0, TAG, MPI_COMM_WORLD);
    MPI_Send(hostname, 80, MPI_CHAR, 0, TAG, MPI_COMM_WORLD);

}

//******************************************************************
// Get important info from master and then setup our world
//
// returns 1 if this slave is needed for the simulation
// return 0 if we don't need the slave (more slaves than rows)
//******************************************************************
int SlaveMPI::initSim()
{
	MPI_Status status;
	double D[PARAM_SIZE];
	double initConds[9];
	int A[1];
	
	if (DEBUG) cout << "Waiting to get some parameters..." << endl;

	// recv the simulation params and put in our param object
	MPI_Recv(D,PARAM_SIZE,MPI_DOUBLE,0,TAG,MPI_COMM_WORLD,&status);

	MPI_Recv(A, 1, MPI_INT, 0, TAG, MPI_COMM_WORLD, &status);

	// Master will send us a 0 if this processor is not needed
	if (A[0] == 0)
	{
		if (DEBUG) cout << "Master told me I wasn't needed :(" << endl;
		return 0;
	}

	// If we were needed, master will tell us the number of slaves computing

	MPI_Recv(A, 1, MPI_INT, 0, TAG, MPI_COMM_WORLD, &status);
	num_p_computing = A[0];

	if (DEBUG) cout << num_p_computing - 1 << " slaves will be working..." << endl;

	if (DEBUG) cout << "Waiting to find out grid info..." << endl;

	// now find out what rows of the grid will be ours
	// how many columns we have, the max size of our point arrays

	MPI_Recv(initConds,9, MPI_DOUBLE,0,TAG,MPI_COMM_WORLD,&status);

	// build the slaveworld using what we found out

	if (DEBUG) cout << "I've been given rows " << initConds[0] 
	                << " through "  << initConds[1] << endl;

	world = new SlaveWorld((int)initConds[0], (int)initConds[1], 
	                       (int)initConds[2], (int)initConds[3]);

	// now that we have a world instatitated, we can set the params
	world->param.setParams(D);	

	world->createWorld();
	world->theGrid->setNumCellsWide((int)initConds[4]);
	world->theGrid->setNumCellsHigh((int)initConds[5]);
	world->theGrid->setCellSize(initConds[6]);
	world->theGrid->setContainerWidth(initConds[7]);
	world->theGrid->setContainerHeight(initConds[8]);
	
	// okay, now we should be getting a big packet with all of our 
	// particle in it

	// find out how many bytes the next packet will be
	MPI_Recv(A,1,MPI_INT,0,TAG,MPI_COMM_WORLD,&status);
	int size = A[0];
	double buffer[size];

	if (DEBUG) cout << "Slave is expecting " << size << " doubles to come in"
		            << " (free)..." << endl;

	MPI_Recv(buffer,size,MPI_DOUBLE,0,TAG,MPI_COMM_WORLD,&status);

	if (DEBUG) cout << "Slave adding particles to grid..." << endl;

	int j;
	Point apoint;

	for (j=0; j<size; j = j + 3)
	{
		apoint.setX(buffer[j]);
		apoint.setY(buffer[j+1]);

		world->theGrid->addNewFreeParticle(apoint, (int) buffer[j+2]);
	}		

	// now the master will send us our boundary particles

	// find out how many bytes the next packet will be
	MPI_Recv(A,1,MPI_INT,0,TAG,MPI_COMM_WORLD,&status);
	
	size = A[0];
	double buffer2[size];
	
	if (DEBUG) cout << "Slave is expecting " << size 
	                << " doubles to come in (boundary)..." << endl;

	MPI_Recv(buffer2,size,MPI_DOUBLE,0,TAG,MPI_COMM_WORLD,&status);

	if (DEBUG)
		cout << "Slave adding boundary particles to grid..." << endl;

	int count = 0;
	for (j=0; j<size; j = j + 4)
	{
		apoint.setX(buffer2[j]);
		apoint.setY(buffer2[j+1]);

		world->theGrid->addNewBndParticle(count, apoint, 
			(int) buffer2[j+2], (int) buffer2[j+3]);
		count++;
	}		

	// tell our grid how mandy shape points we actually have
	world->theGrid->updateNumShapePts(count);

	// now that we have the particles in the grid, we want to build
	// the cell's linked lists

	world->theGrid->buildGrid();

	// We need to get all of the shapes from Master.
	// First we get the number of shapes.
	MPI_Recv(A,1,MPI_INT,0,TAG,MPI_COMM_WORLD,&status);
	world->theGrid->setNumShapes(A[0]);
	
	//Now we need to get all of the attributes.
	double * all_attributes;
	all_attributes = new double[MAX_ATTRS*A[0]];
 
	MPI_Recv(all_attributes, MAX_ATTRS*A[0], MPI_DOUBLE, 0, TAG, 
	                           MPI_COMM_WORLD, &status);

   	//Now loop over the shapes and init the shapes.
   	for(j = 0; j < A[0]; j++)
       		world->theGrid->addShape(all_attributes+j*MAX_ATTRS);
  
	// receive the starting globalID for our slave world
	MPI_Recv(A,1,MPI_INT,0,TAG,MPI_COMM_WORLD,&status);
	world->setGlobalID(A[0]);

	return 1;
}

//******************************************************************
// The main simulation loop
//******************************************************************
void SlaveMPI::simLoop()
{
	MPI_Status status;

	Timer watch_total;		// used to track total time the slaves spends

	// get the name of our computer
        char hostname[80];

	// take no chances
	for (int i=0; i<80; i++)
		hostname[i] = '\0';

        gethostname(hostname, 80);
	if (DEBUG) cout << "Running on " << hostname << endl;

	if (north_rank == -1)
		world->theGrid->setNorthMost(1);
	else
		world->theGrid->setNorthMost(0);

	if (south_rank == -1)
		world->theGrid->setSouthMost(1);
	else
		world->theGrid->setSouthMost(0);

       watch_total.startTimer();

	if (DEBUG) cout << "Slave is doing first loop" << endl;

	world->firstLoop(our_rank);

	// force an output step right away, for debugging purposes

	if (DEBUG) cout << "Going into main loop..." << endl;

	for (int i=0; i<world->param.numSteps; i++)
	{
                if ((world->timeSteps % UPDATE_EVERY) == 0)
                {
			if (our_rank % 2 == 0)
			{
				sendTo(north_rank);
				recvFrom(south_rank);

				sendTo(south_rank);
				recvFrom(north_rank);
			} else {
				recvFrom(south_rank);
				sendTo(north_rank);

				recvFrom(north_rank);
				sendTo(south_rank);
			}

			// now add these newly recv'd particles to our grid
	
			commitReception();
			
			// switch to our other north/south bound list

			world->theGrid->switchBound();
		
		}


		// see if we need to send an update to master

		if (world->timeSteps > world->param.firstOutput)
			if (world->timeSteps % world->param.outputEvery == 0)
				sendMasterUpdate();
		
		// if this was the last time step, no sense in doing the
		// calculations since they won't be saved by the master
		if (i < world->param.numSteps - 1) 
			world->nextLoop(our_rank);
	}

	sendMasterShutdown();

	// output timing info from this slave

        watch_total.stopTimer();
        cout << endl << endl << "Total slave time (w/o init) = " <<
                watch_total.elapsed << endl << endl;

        delete world;
}

//******************************************************************
// Send particles to the either our north or south neighbor
//******************************************************************
void SlaveMPI::sendTo(int dest)
{
	int num_pts, pos, i, j;
	int fringe_pts = 0;
	int migrant_pts = 0;
	int *migrants;
	int *fringe;

	// handle case when points are going to a neighbor slave

	// we must also be sure that the one north of us isn't one
	// that was told to not participate (too many slaves for the
	// the number of rows). 

	if ((dest!=-1) && (dest < num_p_computing))
	{
		// we need to figure out how many we are sending, we will send
		// both the migrant particles and the entire contents of the 
		// border row with the destination

		// size of the migrants and frringes
		if (dest == north_rank)
		{
			migrant_pts = world->theGrid->getNorthBound(&migrants);
			fringe_pts = world->theGrid->getNorthFringe(&fringe);
		} else {
			migrant_pts = world->theGrid->getSouthBound(&migrants);
			fringe_pts = world->theGrid->getSouthFringe(&fringe);

		}

		num_pts = fringe_pts + migrant_pts;


		// allocate an array to put all the outgoing points
		// we will be sending 3 particles for each actual particle
		// since we have the plus and minus one half arrays in grid as well

		double buffer[num_pts * SMOOTH_SIZE * 3];
		SmoothParticle sp;
		int sp_index;

		pos = 0;
	
		// first message tells them how many to expect
	        MPI_Send(&num_pts, 1, MPI_INT, dest, NUMTAG, MPI_COMM_WORLD);

		if (num_pts>0)
		{
			// loop through each point in the linked list and add to buffer

			// first we will add migrant particles
			for (i=0; i<migrant_pts; i++)
			{
				sp_index = migrants[i];
				
				sp = world->theGrid->getParticle(sp_index);
				sp.getPacket(&buffer[pos]);
				pos += SMOOTH_SIZE;

				// now repeat for the plus and minus one half particle 
				sp = world->theGrid->getPlusHalfParticle(sp_index);
				sp.getPacket(&buffer[pos]);
				pos += SMOOTH_SIZE;

				sp = world->theGrid->getMinusHalfParticle(sp_index);
				sp.getPacket(&buffer[pos]);
				pos += SMOOTH_SIZE;

			}
			// now add the fringe row particles to buffer		

			for (i=0; i<fringe_pts; i++)
			{
				sp_index = fringe[i];

				sp = world->theGrid->getParticle(sp_index);
				sp.getPacket(&buffer[pos]);
				pos += SMOOTH_SIZE;

				// now repeat for the plus and minus one half particle
				sp = world->theGrid->getPlusHalfParticle(sp_index);
				sp.getPacket(&buffer[pos]);
				pos += SMOOTH_SIZE;
	
				sp = world->theGrid->getMinusHalfParticle(sp_index);
				sp.getPacket(&buffer[pos]);
				pos += SMOOTH_SIZE;
	
			}
			// send it off to the processor
			MPI_Send(buffer, pos, MPI_DOUBLE, dest, PTSTAG, MPI_COMM_WORLD);
		}
	}

	// now we need to free up the memory that getBody and getFringe
	// allocated

	if (migrant_pts > 0)
		delete [] migrants;
	if (fringe_pts > 0)
		delete [] fringe;
}

//******************************************************************
// Receive particles from one of our neighbors
//******************************************************************
void SlaveMPI::recvFrom(int source)
{
	int num_pts;
	MPI_Status status;

	if ((source!=-1) && (source < num_p_computing))
	{
		// first message tells them how many to expect
		MPI_Recv(&num_pts,1,MPI_INT,source,NUMTAG,MPI_COMM_WORLD,&status);

		// make sure we initialize these in case we don't get any points 

		if (source == north_rank)
			northNumPts = 0;
		else
			southNumPts = 0;
		
		if (num_pts>0)
		{
			int buff_size = num_pts * SMOOTH_SIZE * 3;

			// allocate memory in the proper buffer and receive it
			// (the buffers are class variables)
			if (source == north_rank)
			{
				northBuffer = new double[buff_size];	
				MPI_Recv(northBuffer,buff_size,MPI_DOUBLE,source,
				         PTSTAG,MPI_COMM_WORLD,&status);
				northNumPts = num_pts;
			}
			else
			{
				southBuffer = new double[buff_size];	
				MPI_Recv(southBuffer,buff_size,MPI_DOUBLE,source,
				          PTSTAG,MPI_COMM_WORLD,&status);
				southNumPts = num_pts;
			}
		}
	}

}

//******************************************************************
// Now that we have received particles from the north and south, we want
// to add them to the grid.  This has to be done after the inter-slave
// communication is finished to avoid bouncing migrant particles back
// and getting redundant particles calculated.  
//******************************************************************
void SlaveMPI::commitReception()
{
	SmoothParticle sp, sp_plus_half, sp_minus_half;
	int pos = 0;
	double *buffer;
	int num_pts, i, j;

	for (int buff = 0; buff<2; buff++)
	{
		pos = 0;

		if (buff==0)
		{
			num_pts = northNumPts;
			buffer = northBuffer;
		} else {
			num_pts = southNumPts;
			buffer = southBuffer;
		}

		for (i=0; i<num_pts; i++)
		{
			// now create a new particle and add to grid
			sp.setPacket(&buffer[pos]);

			pos += SMOOTH_SIZE;
		
			// repeat for plus and minus one half parts
			sp_plus_half.setPacket(&buffer[pos]);
			pos += SMOOTH_SIZE;
	
			sp_minus_half.setPacket(&buffer[pos]);
			pos += SMOOTH_SIZE;
	
			world->theGrid->addNewMigrantParticle(sp, sp_plus_half, 
			                                          sp_minus_half);
		}
	}

	// now free up the memory for the buffers

	if (northBuffer != NULL)
		delete [] northBuffer;
	if (southBuffer != NULL)
		delete [] southBuffer;

	northNumPts = 0;
	southNumPts = 0;
	northBuffer = NULL; 
	southBuffer = NULL; 
}

//******************************************************************
// Send the master an update of all the free particles on this processor
//******************************************************************
void SlaveMPI::sendMasterUpdate()
{
	int num_pts, pos, i, j, n;
	int *body_pts;

	num_pts = 0;

	// Get the indices of all the particles in our body
	// This will also return indices of pts in the north or south
	// bound lists (if we do slave updates every time step these
	// will be empty).

	num_pts = world->theGrid->getBody(&body_pts);

	// allocate an array to put all the outgoing points
	// we will only be sending the current particle (not plus 
	// or minus 1/2)

	double buffer[num_pts * SMOOTH_SIZE];
	SmoothParticle sp;
	int sp_index;

	pos = 0;

	if (DEBUG) cout << world->timeSteps << " Telling master to expect " << num_pts 
		            << " particles" << endl;
	
	// first message tells them how many to expect

	MPI_Send(&num_pts, 1, MPI_INT, 0, updateTag, MPI_COMM_WORLD);

	updateTag++;		// we use a unique tag number for each update

	if (num_pts>0)
	{
		// loop through each point in the linked list and add to buffer
		for (i=0; i<num_pts; i++)
		{
			sp_index = body_pts[i];

			sp = world->theGrid->getParticle(sp_index);

			sp.getPacket(&buffer[pos]);
			pos += SMOOTH_SIZE;
		}
	
		// send it off to the master
		MPI_Send(buffer, pos, MPI_DOUBLE, 0, PTS_UPDATE, MPI_COMM_WORLD);
	}	

	// now free up the memory allocated by getBody()

	if (num_pts > 0)
		delete [] body_pts;
}



//******************************************************************
// Tell the master that we are ready to stop the simulation
// We do this by sending it a message indicating -1 points are being sent
//******************************************************************

void SlaveMPI::sendMasterShutdown()
{
	if (DEBUG) cout << "Telling master to shutdown... " << endl;

	int A[1];
	A[0] = -1;

	MPI_Send(A, 1, MPI_INT, 0, updateTag, MPI_COMM_WORLD);

	updateTag++;
}

