#include "World.h"
#include <stdio.h>

//******************************************************************
// Default constructor.
//******************************************************************
World::World() : param(), globalID(0), outFileNum(0),
containerWidth(0), containerHeight(0)
{
	theGrid = new Grid();
	assert(theGrid != NULL);
}


//******************************************************************
// Destructor
//******************************************************************
World::~World()
{
      delete theGrid;
}

//******************************************************************
// Calculates the spacing between particles.
//******************************************************************
void World::calcParticleSpacing()
{
 	double totalArea = containerWidth * containerHeight;
	param.initPtSpacing = sqrt(totalArea / param.numPts);
}

//******************************************************************
// Initialize all fluid and SPH values.
//******************************************************************
void World::calcSPHInitConditions()
{
	param.initMass = param.initDensity * param.initPtSpacing 
	                                   * param.initPtSpacing;
	// Takeda 1994
	param.h = 1.55 * param.initPtSpacing;
	param.betaMax = 1.5;
	
	double tempC1, tempC2, tempC3;
	tempC1 = (param.velocityScale * param.velocityScale) / param.densityVariation;
	tempC2 = (param.nu * param.velocityScale) / 
			(param.lengthScale * param.densityVariation);
	tempC3 = (sqrt(param.bodyFX*param.bodyFX + param.bodyFY*param.bodyFY) * param.lengthScale) / 
					param.densityVariation;
	
	if (tempC1 > tempC2) 
		param.c = tempC1;
	else
		param.c = tempC2;
	
	if (tempC3 > param.c) 
		param.c = tempC3;

// testing new c, 6/10
//	param.c = sqrt(param.c);
	param.c = 5.77E-4;


	double dt1, dt2;
	dt1 = 0.25 * param.h / param.c;
	dt2 = 0.125 * param.h * param.h / param.nu;
}




//******************************************************************
// Calculatyes the initial positions of the free SPH particles so 
// that they fill up the entire Grid. (The grid contains the viewing
// canvas plus some additional buffer room).
//******************************************************************
void World::calcInitSPHPositions()
{
	int i, j;
	int bndCt = 0;
	int freeCt = 0;
	double x, y;
	int shapeID[param.numPts];
	Point freePos[param.numPts];
	Point bndPos[param.numPts];

	// Create list of free position points on the canvas

	for(x = param.initPtSpacing / 2; x < containerWidth;
		           x += param.initPtSpacing)
	{
		for(y = param.initPtSpacing / 2; y < containerHeight; 
			             y += param.initPtSpacing)	
		{

			int n=0;
			while (n < theGrid->getNumShapes())
			{
				if (theGrid->getShape(n)->isInside(x,y))
				{
					bndPos[bndCt].setXY(x,y);
					shapeID[bndCt] = n;
					bndCt++;
					break;
				}
				n++;
			}
			
			if (n == theGrid->getNumShapes())
			{
				freePos[freeCt].setXY(x,y);
				freeCt++;
			}
			
		} // end y loop
	} // end x loop

	param.numPts = freeCt;
	
	// Let theGrid know how many points we have in our Grid so 
	// that it can allocate enough room.

	// We multiply by 4 to make sure we never run out, this
	// could probably be less, but we tested it like this and
	// it works, so we are leaving it alone!

	theGrid->setNumPts(param.numPts * 4);

	// Add each point to the grid and remove the array of points
	// since world doesn't keep track of the partciles.
	for(i = 0; i < param.numPts; i++)
	{ 
		theGrid->addNewFreeParticle(freePos[i], globalID);
		globalID++;
	}

	// Let theGrid know how many points we have in our Grid so 
	// that it can allocate enough room.
	// We need extra space for particles that get pumped in
	// before particles get pushed out. 
	theGrid->setNumShapePts(bndCt);

	// Add each point to the grid and remove the array of points
	// since world doesn't keep track of the partciles.

	for(i = 0; i < bndCt; i++)
	{ 
		theGrid->addNewBndParticle(i, bndPos[i], globalID, shapeID[i]);
		globalID++;
	}


}
	



//******************************************************************
// This intializes everything needed to start our simulation. This
// includes figuring out initial spacing of pts, initializing all
// fluid and SPH values, building the shapes, creating the Grid, 
// initializing positions of SPH particles, initializing the cell 
// size, calculating the number of cells in our Grid, and finally
// building the Grid. 
//******************************************************************
void World::createWorld()
{
	// initial spacing of pts
	calcParticleSpacing();

	// Init fluid and SPH values
	calcSPHInitConditions();

	theGrid->setParam(param);

	theGrid->calcCellSize();
	theGrid->setContainerWidth(containerWidth);
	theGrid->setContainerHeight(containerHeight);

	theGrid->calcNumCellsWideHigh(containerWidth, containerHeight);
	theGrid->setRangeMaster();

	//initializing positions of SPH particles
	calcInitSPHPositions();
	theGrid->buildGrid();

}


//******************************************************************
// Write to file with the particle IDs
// Files of this format can be read by the compare.java and trace.java
// utilities.
//******************************************************************
void World::writeFileDebug()
{
	ofstream fout;
	char file[80];

	if (outFileNum<10)
		sprintf(file, "out.000%d.debug", outFileNum);
	else if (outFileNum<100)
		sprintf(file, "out.00%d.debug", outFileNum);
	else if (outFileNum<1000)
		sprintf(file, "out.0%d.debug", outFileNum);
	else sprintf(file, "out.%d.debug", outFileNum);
		
	fout.open(file);

	(*theGrid).writeID(fout);
	fout.close();

}

//******************************************************************
//******************************************************************
void World::writeFileNeighbor()
{
	ofstream fout;
	char file[80];

	if (outFileNum<10)
		sprintf(file, "out.000%d.neighbor", outFileNum);
	else if (outFileNum<100)
		sprintf(file, "out.00%d.neighbor", outFileNum);
	else if (outFileNum<1000)
		sprintf(file, "out.0%d.neighbor", outFileNum);
	else sprintf(file, "out.%d.neighbor", outFileNum);
		
	fout.open(file);

	(*theGrid).writeNeighbor(fout);
	fout.close();

}

//******************************************************************
// write to an sgi image file 
//******************************************************************
void World::writeFileImage()
{
	ofstream fout;
	char file[80];

	// determine the filename for this image

	if (outFileNum<10)
		sprintf(file, "out.000%d.rgb", outFileNum);
	else if (outFileNum<100)
		sprintf(file, "out.00%d.rgb", outFileNum);
	else if (outFileNum<1000)
		sprintf(file, "out.0%d.rgb", outFileNum);
	else sprintf(file, "out.%d.rgb", outFileNum);
		
	// create an image object and add all the particles to it

	ImageSGI sgi(IMAGE_WIDTH, IMAGE_HEIGHT, SGI_MIN_X, SGI_MIN_Y, 
				SGI_MAX_X, SGI_MAX_Y);

	if (CHECKERS) 
	{
		// draw a grid background

		int segmentsX = IMAGE_WIDTH * 2;
		int segmentsY = IMAGE_HEIGHT * 2;

		double deltaX = (SGI_MAX_X - SGI_MIN_X) / segmentsX;
		double deltaY = (SGI_MAX_Y - SGI_MIN_Y) / segmentsY;

		int hash_i, hash_j;

		for (int i = 0; i < segmentsX; i++)
			for (int j = 0; j < segmentsY; j++)
			{
				hash_i = theGrid->hashI((double) i * deltaX + SGI_MIN_X);
				hash_j = theGrid->hashJ((double) j * deltaY + SGI_MIN_Y);

				if (!((hash_i < 0) || (hash_j <0) ||
				   (hash_i >= theGrid->getNumCellsWide()) ||
				   (hash_j >= theGrid->getNumCellsHigh())))
				{
					if ((hash_i + hash_j) % 2 == 0)
						sgi.add((double) i * deltaX + SGI_MIN_X,
							(double) j * deltaY + SGI_MIN_Y, 32);
					else
						sgi.add((double) i * deltaX + SGI_MIN_X,
							(double) j * deltaY + SGI_MIN_Y, 64);
				}
			}
	}

	SmoothParticle sp;

	for (int i=0; i<theGrid->getNumPts(); i++)
	{
		sp = theGrid->getParticle(i);

		if (!sp.isEmpty())
			sgi.add(sp.getPositionX(), sp.getPositionY(), 255);
	}

	for (int i=0; i<theGrid->getNumShapePts(); i++)
	{
		sp = theGrid->getShapeParticle(i);

		sgi.add(sp.getPositionX(), sp.getPositionY(), 128);

	}

	sgi.write(file);

}

//******************************************************************
// Write out the data in frame format (binary), this can be read
// by the java client application at a later time
//******************************************************************

void World::writeFileClient()
{
	float *buffer;
	int i, j;
	int size = theGrid->getFrame(&buffer);

	ofstream fout;
	char file[80];

	// determine the filename for this image

	if (outFileNum<10)
		sprintf(file, "out.000%d.dat", outFileNum);
	else if (outFileNum<100)
		sprintf(file, "out.00%d.dat", outFileNum);
	else if (outFileNum<1000)
		sprintf(file, "out.0%d.dat", outFileNum);
	else sprintf(file, "out.%d.rgb", outFileNum);

	fout.open(file);		

	// On some types of systems (linux) we need to reverse the
	// order of our bytes in the float.  Other systems don't (sunOS)

	if (REV_OUTPUT_CLIENT)
	{
		char tempbuffer[sizeof(float)];
	
		char *buff = (char *) buffer;

                // we need to reverse the order of each set of bytes
                for (i = 0; i < size; i++)
                {
			// copy the bytes out of the original buffer
                        for (j=0; j<sizeof(float); j++)
				tempbuffer[j] = buff[i * sizeof(float) + j];

			// put them back in reverse order

                        for (j=0; j<sizeof(float); j++)
				buff[i * sizeof(float) + j] = 
					tempbuffer[sizeof(float) - j - 1];		
                }		
	}
	
	fout.write( (char *) buffer, size * sizeof(float));

	fout.close();
}

//******************************************************************
//******************************************************************
void World::incrementTime()
{
	outFileNum++;
}

